# pyright: reportIncompatibleMethodOverride=false

from collections.abc import Iterable, Set as AbstractSet
from typing import Any, Literal, Never, Self, final
from typing_extensions import TypeAliasType, TypeVar, override

_AnySet = TypeAliasType("_AnySet", AbstractSet[Any])
_SetT = TypeVar("_SetT", bound=_AnySet)

@final
class EmptyType(AbstractSet[Never]):
    @staticmethod
    def get() -> EmptyType: ...
    def __new__(cls: Never, /) -> Never: ...  # raises TypeError
    def __init__(self: Never, /) -> Never: ...  # raises TypeError
    def __bool__(self, /) -> Literal[False]: ...
    @override
    def __len__(self, /) -> Literal[0]: ...
    @override
    def __iter__(self, /) -> Self: ...
    def __next__(self, /) -> Never: ...
    @override
    def __contains__(self, x: object, /) -> Literal[False]: ...
    @override
    def __hash__(self, /) -> int: ...
    @override
    def __eq__(self, other: object, /) -> bool: ...
    @override
    def __ne__(self, other: object, /) -> bool: ...
    @override
    def __lt__(self, other: _AnySet, /) -> bool: ...
    @override
    def __le__(self, other: _AnySet, /) -> Literal[True]: ...
    @override
    def __gt__(self, other: _AnySet, /) -> Literal[False]: ...
    @override
    def __ge__(self, other: _AnySet, /) -> bool: ...
    @override
    def __and__(self, other: _AnySet, /) -> Self: ...
    @override
    def __sub__(self, other: _AnySet, /) -> Self: ...
    def __rsub__(self, other: _SetT, /) -> _SetT: ...
    @override
    def __or__(self, other: _SetT, /) -> _SetT: ...
    @override
    def __xor__(self, other: _SetT, /) -> _SetT: ...
    @override
    def isdisjoint(self, other: Iterable[Any], /) -> Literal[True]: ...
    @property
    def C(self, /) -> UniverseType: ...  # noqa: N802

@final
class UniverseType(AbstractSet[object]):
    @staticmethod
    def get() -> UniverseType: ...
    def __new__(cls: Never, /) -> Never: ...  # raises TypeError
    def __init__(self: Never, /) -> Never: ...  # raises TypeError
    def __bool__(self, /) -> Literal[True]: ...
    @override
    def __len__(self, /) -> Never: ...  # raises OverflowError
    @override
    def __iter__(self, /) -> Never: ...  # raises OverflowError
    @override
    def __contains__(self, x: object, /) -> Literal[True]: ...
    @override
    def __hash__(self, /) -> int: ...
    @override
    def __eq__(self, other: object, /) -> bool: ...
    @override
    def __ne__(self, other: object, /) -> bool: ...
    @override
    def __lt__(self, other: _AnySet, /) -> Literal[False]: ...
    @override
    def __le__(self, other: _AnySet, /) -> bool: ...
    @override
    def __gt__(self, other: _AnySet, /) -> bool: ...
    @override
    def __ge__(self, other: _AnySet, /) -> Literal[True]: ...
    @override
    def __and__(self, other: _SetT, /) -> _SetT: ...
    @override
    def __sub__(self, other: _AnySet, /) -> Self | EmptyType: ...  # might raise
    def __rsub__(self, other: _AnySet, /) -> EmptyType: ...
    @override
    def __or__(self, other: _AnySet, /) -> Self: ...
    @override
    def __xor__(self, other: _SetT, /) -> _SetT | EmptyType: ...
    @override
    def isdisjoint(self, other: Iterable[Any], /) -> bool: ...  # raises if not a set
    @property
    def C(self, /) -> EmptyType: ...  # noqa: N802



import pytest


def test_import_symset() -> None:
    try:
        import symset  # noqa: F401, PLC0415
    except Exception as e:  # noqa: BLE001
        pytest.fail(f"Failed to import symset: {e}")



"""Symbolic sets."""

from collections.abc import Set as AbstractSet
from typing import Final

from ._core import EmptyType, UniverseType

AbstractSet.register(EmptyType)  # pyright: ignore[reportAttributeAccessIssue, reportUnknownMemberType]
AbstractSet.register(UniverseType)  # pyright: ignore[reportAttributeAccessIssue, reportUnknownMemberType]
del AbstractSet


Empty: Final = EmptyType.get()
Universe: Final = UniverseType.get()


__all__ = "Empty", "Universe"


def __dir__() -> tuple[str, ...]:
    return __all__



# pyright: reportUnknownVariableType=false

from collections.abc import Set as AbstractSet
from typing import Final, Never

import pytest
from hypothesis import given, strategies as st

from symset import Empty, EmptyType, Universe

_EMPTY_BUILTIN_SET: Final[set[Never]] = set()
_EMPTY_FROZENSET: Final[frozenset[Never]] = frozenset(())
_FALSY_NON_SET: Final[tuple[object, ...]] = None, False, 0, 0.0, "", b"", (), [], {}


def test_subclass_abc() -> None:
    assert isinstance(Empty, AbstractSet)
    assert issubclass(EmptyType, AbstractSet)


def test_cannot_construct() -> None:
    with pytest.raises(TypeError):
        _ = EmptyType()  # pyright: ignore[reportGeneralTypeIssues]


def test_no_dict() -> None:
    assert not hasattr(Empty, "__dict__")


def test_no_pyo3_internals() -> None:
    assert not hasattr(Empty, "__richcmp__")
    assert not hasattr(Empty, "__concat__")
    assert not hasattr(Empty, "__repeat__")
    assert not hasattr(Empty, "__traverse__")
    assert not hasattr(Empty, "__clear__")


def test_repr() -> None:
    assert repr(Empty) == "Empty"


def test_str() -> None:
    assert str(Empty) == "∅"


def test_bool() -> None:
    assert not Empty


def test_len() -> None:
    assert len(Empty) == 0


@given(st.none() | st.booleans() | st.integers() | st.floats() | st.text())
def test_contains(value: float | str | None) -> None:
    assert value not in Empty


def test_iter() -> None:
    assert sum(1 for _ in iter(Empty)) == 0


def test_hash() -> None:
    assert {Empty} == {_EMPTY_FROZENSET}
    assert hash(Empty) == hash(_EMPTY_FROZENSET)
    assert Empty._hash() == hash(_EMPTY_FROZENSET)  # noqa: SLF001


@pytest.mark.parametrize("other", [Empty, _EMPTY_FROZENSET, _EMPTY_BUILTIN_SET])
def test_eq(other: object) -> None:
    assert Empty == other


@pytest.mark.parametrize("other", [*_FALSY_NON_SET, {None}, frozenset({None})])
def test_ne(other: object) -> None:
    assert Empty != other


def test_lt() -> None:
    assert not Empty < Empty
    assert not Empty < _EMPTY_FROZENSET
    assert not Empty < _EMPTY_BUILTIN_SET

    assert Empty < {object()}
    assert Empty < frozenset({object()})

    with pytest.raises(TypeError):
        _ = Empty < ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty < []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty < {}  # pyright: ignore[reportOperatorIssue]


def test_le() -> None:
    assert Empty <= Empty
    assert Empty <= _EMPTY_FROZENSET
    assert Empty <= _EMPTY_BUILTIN_SET

    assert Empty <= {object()}
    assert Empty <= frozenset({object()})

    with pytest.raises(TypeError):
        _ = Empty <= ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty <= []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty <= {}  # pyright: ignore[reportOperatorIssue]


def test_ge() -> None:
    assert Empty >= Empty
    assert Empty >= _EMPTY_FROZENSET
    assert Empty >= _EMPTY_BUILTIN_SET

    assert not Empty >= {object()}
    assert not Empty >= frozenset({object()})

    with pytest.raises(TypeError):
        _ = Empty >= ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty >= []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty >= {}  # pyright: ignore[reportOperatorIssue]


def test_gt() -> None:
    assert not Empty > Empty
    assert not Empty > _EMPTY_FROZENSET
    assert not Empty > _EMPTY_BUILTIN_SET

    assert not Empty > {object()}
    assert not Empty > frozenset({object()})

    with pytest.raises(TypeError):
        _ = Empty > ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty > []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty > {}  # pyright: ignore[reportOperatorIssue]


def test_and() -> None:
    assert Empty & Empty is Empty
    assert Empty & _EMPTY_FROZENSET is Empty
    assert Empty & _EMPTY_BUILTIN_SET is Empty

    assert Empty & {object()} is Empty
    assert Empty & frozenset({object()}) is Empty

    with pytest.raises(TypeError):
        _ = Empty & ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty & []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty & {}  # pyright: ignore[reportOperatorIssue]


def test_or() -> None:
    assert Empty | Empty is Empty
    assert Empty | _EMPTY_FROZENSET is Empty
    assert Empty | _EMPTY_BUILTIN_SET is Empty

    s1 = {object()}
    f1 = frozenset({object()})
    assert Empty | s1 is s1
    assert Empty | f1 is f1

    with pytest.raises(TypeError):
        _ = Empty | ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty | []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty | {}  # pyright: ignore[reportOperatorIssue]


def test_xor() -> None:
    assert Empty ^ Empty is Empty
    assert Empty ^ _EMPTY_FROZENSET is Empty
    assert Empty ^ _EMPTY_BUILTIN_SET is Empty

    s1 = {object()}
    f1 = frozenset({object()})
    assert Empty ^ s1 is s1
    assert Empty ^ f1 is f1

    with pytest.raises(TypeError):
        _ = Empty ^ ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty ^ []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty ^ {}  # pyright: ignore[reportOperatorIssue]


def test_sub() -> None:
    assert Empty - Empty is Empty
    assert Empty - _EMPTY_FROZENSET is Empty
    assert Empty - _EMPTY_BUILTIN_SET is Empty

    assert Empty - {object()} is Empty
    assert Empty - frozenset({object()}) is Empty

    with pytest.raises(TypeError):
        _ = Empty - ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty - []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Empty - {}  # pyright: ignore[reportOperatorIssue]


def test_rsub() -> None:
    assert Empty - Empty is Empty
    assert _EMPTY_FROZENSET - Empty is Empty
    assert _EMPTY_BUILTIN_SET - Empty is Empty

    s1 = {object()}
    f1 = frozenset({object()})
    assert s1 - Empty is s1
    assert f1 - Empty is f1

    with pytest.raises(TypeError):
        _ = () - Empty  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = [] - Empty  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = {} - Empty  # pyright: ignore[reportOperatorIssue]


def test_isdisjoint() -> None:
    assert Empty.isdisjoint(Empty)
    assert Empty.isdisjoint(_EMPTY_FROZENSET)
    assert Empty.isdisjoint(_EMPTY_BUILTIN_SET)

    assert Empty.isdisjoint({object()})
    assert Empty.isdisjoint(frozenset({object()}))

    with pytest.raises(TypeError):
        _ = Empty.isdisjoint(None)  # pyright: ignore[reportArgumentType]
    with pytest.raises(TypeError):
        _ = Empty.isdisjoint(object())  # pyright: ignore[reportArgumentType]
    with pytest.raises(TypeError):
        _ = Empty.isdisjoint(EmptyType)  # pyright: ignore[reportArgumentType]


def test_complement() -> None:
    assert Empty.C is Universe, (Empty.C, Universe)



# pyright: reportUnknownVariableType=false

from collections.abc import Set as AbstractSet
from typing import Final, Never

import pytest
from hypothesis import given, strategies as st

from symset import Empty, EmptyType, Universe, UniverseType

_EMPTY_BUILTIN_SET: Final[set[Never]] = set()
_EMPTY_FROZENSET: Final[frozenset[Never]] = frozenset(())

_FALSY_NON_SET: Final[tuple[object, ...]] = None, False, 0, 0.0, "", b"", (), [], {}


def test_subclass_abc() -> None:
    assert isinstance(Universe, AbstractSet)
    assert issubclass(UniverseType, AbstractSet)


def test_cannot_construct() -> None:
    with pytest.raises(TypeError):
        _ = UniverseType()  # pyright: ignore[reportGeneralTypeIssues]


def test_no_dict() -> None:
    assert not hasattr(Universe, "__dict__")


def test_no_pyo3_internals() -> None:
    assert not hasattr(Universe, "__richcmp__")
    assert not hasattr(Universe, "__concat__")
    assert not hasattr(Universe, "__repeat__")
    assert not hasattr(Universe, "__traverse__")
    assert not hasattr(Universe, "__clear__")


def test_repr() -> None:
    assert repr(Universe) == "Universe"


def test_str() -> None:
    assert str(Universe) == "U"


def test_bool() -> None:
    assert Universe


def test_len() -> None:
    with pytest.raises(OverflowError):
        _ = len(Universe)


@given(st.none() | st.booleans() | st.integers() | st.floats() | st.text())
def test_contains_value(value: float | str | None) -> None:
    assert value in Universe


@pytest.mark.parametrize("value", [Empty, Universe])
def test_contains_special_set(value: EmptyType | UniverseType) -> None:
    assert value in Universe


def test_iter() -> None:
    with pytest.raises(OverflowError):
        _ = iter(Universe)


def test_hash() -> None:
    assert hash(Universe) == -hash(_EMPTY_FROZENSET) - 1


def test_eq() -> None:
    assert Universe is Universe


@pytest.mark.parametrize(
    "other", [*_FALSY_NON_SET, _EMPTY_FROZENSET, _EMPTY_BUILTIN_SET, Empty]
)
def test_ne(other: object) -> None:
    assert Universe != other


def test_lt() -> None:
    assert not Universe < Universe
    assert not Universe < Empty
    assert not Universe < _EMPTY_FROZENSET
    assert not Universe < _EMPTY_BUILTIN_SET

    assert not Universe < {object()}
    assert not Universe < frozenset({object()})

    with pytest.raises(TypeError):
        _ = Universe < ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe < []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe < {}  # pyright: ignore[reportOperatorIssue]


def test_le() -> None:
    assert Universe <= Universe
    assert not Universe <= Empty
    assert not Universe <= _EMPTY_FROZENSET
    assert not Universe <= _EMPTY_BUILTIN_SET
    assert not Universe <= {object()}
    assert not Universe <= frozenset({object()})

    with pytest.raises(TypeError):
        _ = Universe <= ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe <= []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe <= {}  # pyright: ignore[reportOperatorIssue]


def test_ge() -> None:
    assert Universe >= Universe
    assert Universe >= Empty
    assert Universe >= _EMPTY_FROZENSET
    assert Universe >= _EMPTY_BUILTIN_SET
    assert Universe >= {object()}
    assert Universe >= frozenset({object()})

    with pytest.raises(TypeError):
        _ = Universe >= ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe >= []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe >= {}  # pyright: ignore[reportOperatorIssue]


def test_gt() -> None:
    assert not Universe > Universe
    assert Universe > Empty
    assert Universe > _EMPTY_FROZENSET
    assert Universe > _EMPTY_BUILTIN_SET
    assert Universe > {object()}
    assert Universe > frozenset({object()})

    with pytest.raises(TypeError):
        _ = Universe > ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe > []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe > {}  # pyright: ignore[reportOperatorIssue]


def test_and() -> None:
    assert Universe & Universe is Universe
    assert Universe & Empty is Empty
    assert Universe & _EMPTY_FROZENSET is Empty
    assert Universe & _EMPTY_BUILTIN_SET is Empty

    s1 = {object()}
    f1 = frozenset({object()})
    assert Universe & s1 is s1
    assert Universe & f1 is f1

    with pytest.raises(TypeError):
        _ = Universe & ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe & []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe & {}  # pyright: ignore[reportOperatorIssue]


def test_or() -> None:
    assert Universe | Universe is Universe
    assert Universe | Empty is Universe
    assert Universe | _EMPTY_FROZENSET is Universe
    assert Universe | _EMPTY_BUILTIN_SET is Universe

    s1 = {object()}
    f1 = frozenset({object()})
    assert Universe | s1 is Universe
    assert Universe | f1 is Universe

    with pytest.raises(TypeError):
        _ = Universe | ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe | []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe | {}  # pyright: ignore[reportOperatorIssue]


def test_xor() -> None:
    assert Universe ^ Universe is Empty
    assert Universe ^ Empty is Universe
    assert Universe ^ _EMPTY_FROZENSET is Universe
    assert Universe ^ _EMPTY_BUILTIN_SET is Universe

    # TODO: absolute complement
    with pytest.raises(NotImplementedError):
        _ = Universe ^ {object()}

    with pytest.raises(TypeError):
        _ = Universe ^ ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe ^ []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe ^ {}  # pyright: ignore[reportOperatorIssue]


def test_sub() -> None:
    assert Universe - Universe is Empty
    assert Universe - Empty is Universe
    assert Universe - _EMPTY_FROZENSET is Universe
    assert Universe - _EMPTY_BUILTIN_SET is Universe

    # TODO: absolute complement
    with pytest.raises(NotImplementedError):
        _ = Universe - {object()}

    with pytest.raises(TypeError):
        _ = Universe - ()  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe - []  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = Universe - {}  # pyright: ignore[reportOperatorIssue]


def test_rsub() -> None:
    assert _EMPTY_FROZENSET - Universe is Empty
    assert _EMPTY_BUILTIN_SET - Universe is Empty

    s1 = {object()}
    f1 = frozenset({object()})
    assert s1 - Universe is Empty
    assert f1 - Universe is Empty

    with pytest.raises(TypeError):
        _ = () - Universe  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = [] - Universe  # pyright: ignore[reportOperatorIssue]
    with pytest.raises(TypeError):
        _ = {} - Universe  # pyright: ignore[reportOperatorIssue]


def test_isdisjoint() -> None:
    assert not Universe.isdisjoint(Universe)
    assert Universe.isdisjoint(Empty)
    assert Universe.isdisjoint(_EMPTY_FROZENSET)
    assert Universe.isdisjoint(_EMPTY_BUILTIN_SET)

    assert not Universe.isdisjoint({object()})
    assert not Universe.isdisjoint(frozenset({object()}))

    with pytest.raises(TypeError):
        _ = Universe.isdisjoint(None)  # pyright: ignore[reportArgumentType]
    with pytest.raises(TypeError):
        _ = Universe.isdisjoint(object())  # pyright: ignore[reportArgumentType]
    with pytest.raises(TypeError):
        _ = Universe.isdisjoint(EmptyType)  # pyright: ignore[reportArgumentType]


def test_complement() -> None:
    assert Universe.C is Empty, (Universe.C, Empty)
